#include <torch/extension.h>
#include <vector>
#include <cfloat>
#include <cuda_runtime.h>

// Custom fused CUDA kernel that performs max pooling, global average pooling, bias addition, and an optional reduction (sum) along a given dimension.
// This kernel minimizes __syncthreads usage by using warp-level primitives for intra-warp reduction.

// Each block handles one (n, c) slice of the 5D tensor (N, C, D, H, W) output from conv3d.
// We assume that the spatial dimensions D, H, and W are exactly divisible by the pooling sizes.

__global__ void fusedPoolingAndReductionKernel(
    const float* __restrict__ conv_out,  // Input from conv3d, shape: (N, C, D, H, W)
    int N,
    int C,
    int D,
    int H,
    int W,
    int poolD,
    int poolH,
    int poolW,
    const float* __restrict__ bias,      // Bias tensor, assumed shape: (C) (broadcasted along N)
    int sum_dim,                         // If sum_dim==1, we sum over the channel dimension
    float* __restrict__ output,          // Output tensor: if sum_dim==1, shape: (N); else shape: (N, C)
    float divisor                        // Division factor applied per conv output element
) {
    // Determine which (n, c) slice this block is responsible for
    int nc = blockIdx.x; // block index in [0, N*C)
    int n = nc / C;
    int c = nc % C;

    // Dimensions after max pooling (using non-overlapping windows with size equal to pool dimensions)
    int outD = D / poolD;
    int outH = H / poolH;
    int outW = W / poolW;
    int numPools = outD * outH * outW;
    const float invDiv = 1.0f / divisor;

    float thread_sum = 0.0f;

    // Each thread processes multiple pooling windows in a strided loop
    for (int idx = threadIdx.x; idx < numPools; idx += blockDim.x) {
        // Calculate the pooling window indices in the output of max pooling
        int d = idx / (outH * outW);
        int rem = idx % (outH * outW);
        int h = rem / outW;
        int w = rem % outW;

        int startD = d * poolD;
        int startH = h * poolH;
        int startW = w * poolW;

        // Compute max pooling over the window
        float max_val = -FLT_MAX;
        #pragma unroll
        for (int pd = 0; pd < poolD; pd++) {
            #pragma unroll
            for (int ph = 0; ph < poolH; ph++) {
                #pragma unroll
                for (int pw = 0; pw < poolW; pw++) {
                    int curD = startD + pd;
                    int curH = startH + ph;
                    int curW = startW + pw;
                    // Compute linear index for conv_out: shape is (N, C, D, H, W)
                    int index = (((n * C + c) * D + curD) * H + curH) * W + curW;
                    // Incorporate division on the fly for better fusion
                    float val = __fmaf_rn(__ldg(&conv_out[index]), invDiv, 0.0f);
                    if (val > max_val) {
                        max_val = val;
                    }
                }
            }
        }
        // Accumulate the max via this pooling window
        thread_sum += max_val;
    }

    // Use warp-level reduction to minimize __syncthreads calls
    unsigned int lane = threadIdx.x & 31;
    // Intra-warp reduction using shuffle
    float sum_val = thread_sum;
    for (int offset = 16; offset > 0; offset /= 2) {
        sum_val += __shfl_down_sync(0xFFFFFFFF, sum_val, offset);
    }

    // Allocate shared memory for inter-warp reduction; only one __syncthreads call is used
    extern __shared__ float shared[];
    if (lane == 0) {
        shared[threadIdx.x >> 5] = sum_val;
    }
    __syncthreads();

    // First warp reduces the partial sums from each warp
    float block_sum = 0.0f;
    int numWarps = (blockDim.x + 31) >> 5;
    if (threadIdx.x < numWarps) {
        block_sum = shared[threadIdx.x];
    }
    if (threadIdx.x < 32) { // reduction within first warp
        for (int offset = 16; offset > 0; offset /= 2) {
            block_sum += __shfl_down_sync(0xFFFFFFFF, block_sum, offset);
        }
    }

    // Thread 0 of the block finalizes the result for this (n,c) slice
    if (threadIdx.x == 0) {
        // Compute global average from the max pooling results
        float avg = block_sum / numPools;
        // Add the bias (broadcasted along n)
        avg = avg + bias[c];

        // Write out the result. If sum_dim == 1, accumulate into output for each n; otherwise, store per (n,c).
        if (sum_dim == 1) {
            atomicAdd(&output[n], avg);
        } else {
            output[n * C + c] = avg;
        }
    }
}


// The unified forward_cuda function that calls conv3d and then the fused pooling/reduction kernel

torch::Tensor forward_cuda(
    torch::Tensor x,
    double divisor,
    std::vector<int64_t> pool_size,  // [poolD, poolH, poolW]
    int64_t sum_dim,
    torch::Tensor conv_weight,
    torch::Tensor conv_bias,
    torch::Tensor bias
) {
    // Check that inputs are CUDA tensors
    TORCH_CHECK(x.is_cuda(), "x must be a CUDA tensor.");
    TORCH_CHECK(conv_weight.is_cuda(), "conv_weight must be a CUDA tensor.");
    TORCH_CHECK(conv_bias.is_cuda(), "conv_bias must be a CUDA tensor.");
    TORCH_CHECK(bias.is_cuda(), "bias must be a CUDA tensor.");

    // 1) 3D convolution (we do not fuse convolution here as cuDNN is highly optimized)
    auto conv_out = at::conv3d(x, conv_weight, conv_bias);
    // conv_out shape: (N, C, D, H, W)
    int N = conv_out.size(0);
    int C = conv_out.size(1);
    int D = conv_out.size(2);
    int H = conv_out.size(3);
    int W = conv_out.size(4);

    // 2) Launch the fused pooling and reduction kernel
    // The kernel will perform: max pooling (with window size = pool_size), global average pooling, bias addition and sum reduction (if requested)
    // Each block processes one (n, c) slice
    int totalBlocks = N * C;
    int threadsPerBlock = 128;
    int sharedMemSize = ((threadsPerBlock + 31) / 32) * sizeof(float);

    // Prepare output tensor
    torch::Tensor output;
    if (sum_dim == 1) {
        // If summing over channels, output will have shape (N) (one value per sample)
        output = torch::zeros({N}, conv_out.options());
    } else {
        // Otherwise, store the result per channel: shape (N, C)
        output = torch::empty({N, C}, conv_out.options());
    }

    fusedPoolingAndReductionKernel<<<totalBlocks, threadsPerBlock, sharedMemSize>>>(
        conv_out.data_ptr<float>(),
        N, C, D, H, W,
        pool_size[0], pool_size[1], pool_size[2],
        bias.data_ptr<float>(),
        sum_dim,
        output.data_ptr<float>(),
        static_cast<float>(divisor)
    );

    cudaDeviceSynchronize();
    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward_cuda, "Fused 3D Conv, Pooling, GlobalAvgPool, BiasAdd, and Sum (CUDA) with optimized synchronization");
}
